library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity VGA is

  port (
    px_clk, rst : in  std_logic;
    R, G, B     : out std_logic_vector (7 downto 0);
    HS, VS      : out std_logic);

end entity VGA;

architecture myVGA of VGA is

  -- Horizontal timings
  constant H_VISIBLE_AREA : integer := 640;
  constant H_FP           : integer := 16;
  constant H_SYNC         : integer := 96;
  constant H_BP           : integer := 48;
  constant H_ALL          : integer := 800;

  -- Vertical timings
  constant V_VISIBLE_AREA : integer := 480;
  constant V_FP           : integer := 10;
  constant V_SYNC         : integer := 2;
  constant V_BP           : integer := 33;
  constant V_ALL          : integer := 525;

  -- local
  signal h_px_count : integer range 0 to H_ALL;
  signal v_px_count : integer range 0 to V_ALL;
  signal can_draw   : std_logic;

  -- ball
  signal ball_x     : integer range 0 to H_VISIBLE_AREA;
  signal ball_y     : integer range 0 to V_VISIBLE_AREA;
  signal color_mask : std_logic_vector (7 downto 0);
  signal BALL_WIDTH : integer range 10 to 50;
  signal ball_r     : integer range 0 to 255;  -- colors
  signal ball_g     : integer range 0 to 255;
  signal ball_b     : integer range 0 to 255;

begin  -- architecture myVGA

  VH_PX_COUNTER : process (px_clk, rst) is
  begin  -- process H_PX_COUNTER
    if rst = '0' then                   -- asynchronous reset (active low)
      h_px_count <= 0;
    elsif rising_edge(px_clk) then      -- rising clock edge
      if h_px_count < H_ALL then
        h_px_count <= h_px_count + 1;
      else
        h_px_count <= 0;
        v_px_count <= v_px_count + 1;
      end if;
      if v_px_count = V_ALL then
        v_px_count <= 0;
      end if;
    end if;
  end process VH_PX_COUNTER;

  V_SYNC_GEN : process (px_clk, rst, v_px_count) is
  begin  -- process V_SYNC_GEN
    if rst = '0' then                   -- asynchronous reset (active low)
      VS <= '1';
    elsif rising_edge(px_clk) then      -- rising clock edge
      if v_px_count >= V_VISIBLE_AREA + V_FP and v_px_count < V_ALL - V_BP then
        VS <= '0';
      else
        VS <= '1';
      end if;
    end if;
  end process V_SYNC_GEN;

  H_SYNC_GEN : process (px_clk, rst, h_px_count) is
  begin  -- process H_SYNC_GEN
    if rst = '0' then                   -- asynchronous reset (active low)
      HS <= '1';
    elsif rising_edge(px_clk) then      -- rising clock edge
      if h_px_count >= H_VISIBLE_AREA + H_FP and h_px_count < H_ALL - H_BP then
        HS <= '0';
      else
        HS <= '1';
      end if;
    end if;
  end process H_SYNC_GEN;

  CAN_DRAW_GEN : process (px_clk, rst, v_px_count, h_px_count) is
  begin  -- process CAN_DRAW_GEN
    if rst = '0' then                   -- asynchronous reset (active low)

    elsif rising_edge(px_clk) then      -- rising clock edge
      if v_px_count < V_VISIBLE_AREA and h_px_count < H_VISIBLE_AREA then
        can_draw <= '1';
      else
        can_draw <= '0';
      end if;
    end if;
  end process CAN_DRAW_GEN;

  process (can_draw) is
  begin  -- process
    if can_draw = '1' then
      R <= color_mask and std_logic_vector(to_unsigned(ball_r, 8));
      G <= color_mask and std_logic_vector(to_unsigned(ball_g, 8));
      B <= color_mask and std_logic_vector(to_unsigned(ball_b, 8));
    else
      R <= "00000000";
      G <= "00000000";
      B <= "00000000";
    end if;
  end process;

  ball_bounce : process (v_px_count) is
    variable count        : integer range 0 to 300;
    variable ball_speed_x : integer range -1 to 1 := 1;
    variable ball_speed_y : integer range -1 to 1 := 1;
  begin  -- process ball_bounce
    if rising_edge(px_clk) and v_px_count = 0 then

      if count < 300 then
        count := count + 1;
      else
        count := 0;

        if ball_x >= H_VISIBLE_AREA - BALL_WIDTH then
          ball_speed_x := - ball_speed_x;
        end if;
        if ball_y >= V_VISIBLE_AREA - BALL_WIDTH then
          ball_speed_y := - ball_speed_y;
        end if;

        if ball_x < BALL_WIDTH then
          ball_x       <= BALL_WIDTH + 1;
          ball_speed_x := abs(ball_speed_x);
        end if;
        if ball_y < BALL_WIDTH then
          ball_y       <= BALL_WIDTH + 1;
          ball_speed_y := abs(ball_speed_y);
        end if;

        ball_x <= ball_x + ball_speed_x;
        ball_y <= ball_y + ball_speed_y;

        -- random colors
        ball_r <= ball_r + 2;
        ball_g <= ball_g + 1;
        ball_b <= ball_b + 3;
      end if;

    end if;
  end process ball_bounce;

  ball_draw : process (px_clk) is
  begin  -- process ball_draw
    if rising_edge(px_clk) then
      if h_px_count < ball_x + BALL_WIDTH and h_px_count >= ball_x - BALL_WIDTH and
        v_px_count < ball_y + BALL_WIDTH and v_px_count >= ball_y - BALL_WIDTH then
        color_mask <= "11111111";
      else
        color_mask <= "00000000";
      end if;
    end if;
  end process ball_draw;

  vary_ball_width : process (v_px_count) is
    variable count  : integer range 0 to 1023;
    variable growth : integer range -1 to 1 := 1;
  begin  -- process ball_bounce
    if rising_edge(px_clk) and v_px_count = 0 then
      count := count + 1;
      if count = 1023 then
        if BALL_WIDTH = 50 then
          growth := -1;
        elsif BALL_WIDTH = 10 then
          growth := 1;
        end if;
        BALL_WIDTH <= BALL_WIDTH + growth;
      end if;
    end if;
  end process;
end architecture myVGA;
